// PackFile.cpp: implementation of the CPackFile class.
//
//////////////////////////////////////////////////////////////////////


#include "PackFile.h"
#include "DataManager.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

namespace FileSystem
{
//pack file id
char id[]="FileSystem Pack File\0";


CResult CPackFile::initialize(char* newFileName)
{

	CResult result=resultOk;

	//initialize the pack file
	result = initialize();

	if (result==resultOk)
	{

		//store filename
		strcpy(fileName,newFileName);

		//load the pack from file
		result = loadFromFile(newFileName);
	}

	return result;
}

CResult CPackFile::initialize(char* directory, char* newFileName, 
							  DWORD blockSize)
{
	CResult result=resultOk;

	//initialize the pack file
	result = initialize();

	//if al is ok
	if (result == resultOk)
	{
		//set vars
		strcpy(fileName,newFileName);
		header.blocksize = blockSize;
		
		///make the pack from directory
		result = makeFromDir(directory);

		//if all its ok
		if (result==resultOk)
		{
			//if we have any file, save pack file
			if (header.numberOfFiles>0)
				result = savePackFile(newFileName);
			else
				//repor error string
				result = CResult(codeFail,0,
							"CPackFile::initialize couldn't find any file.");
			
		}
	}


	//return result
	return result;
}

CResult CPackFile::initialize()
{
	CResult result = resultOk;

	//init vars
	ZeroMemory(&header,sizeof(header));	
	strcpy(header.mark,id);
	header.version = 1;
	header.subversion = 0;
	header.numberOfFiles = 0;
	firstFile = NULL;
	lastFile = NULL;
	

	//call base object
	result = Inherit::initialize();

	//return value
	return result;
}


CResult CPackFile::loadFromFile(char* fileName)
{
	CResult result = resultOk;

	//open input file
	FILE* fileInputStream;
	fileInputStream = fopen(fileName,"rb");

	//if we couldn't open repor error
	if (fileInputStream == NULL)
		result = CResult(codeFail,0,
						  "CPackFile::loadFromFile couldn't open input file");
	else
	{
		//load pack header
		result = loadHeader(fileInputStream);

		//if al is ok load file table
		if (result == resultOk)
			result = loadFileTable(fileInputStream);

		//close the file		
		fclose(fileInputStream);
	}

	//return value
	return result;
}



CResult CPackFile::makeFromDir(char *directory)
{
	CResult result = resultOk;

	//get the current directory
	char basedir[MAX_PATH];	
	ZeroMemory(basedir,sizeof(basedir));	
	DWORD currentDirResult = GetCurrentDirectory(sizeof(basedir),basedir);

	//if we couldn't get the directory, report a error string
	if (currentDirResult==0)
		result = CResult(codeFail,0,
					"CPackFile::makeFromDir couldn't find current directory.");

	else
	{
		//if we use relative path
		if (strstr(directory,":")==NULL)
		{
			//concat dirs
			strcat(basedir,"\\");
			strcat(basedir,directory);
		}
		else
		{
			//just use the one passed
			strcpy(basedir,directory);
		}

		//add the directory and contes
		result = addDirectory(basedir,basedir);
	}

	//return value
	return result;
}

CResult CPackFile::addDirectory(char *directory,char* basedir)
{	
	CResult result = resultOk;

	//set the file search pattern
	char pattern[MAX_PATH];
	ZeroMemory(pattern,sizeof(pattern));	
	strcat(pattern,directory);
	strcat(pattern,"\\*.*");		

	//data to find
	WIN32_FIND_DATA findData;
	
	//find first file
	HANDLE findFileHandle = NULL;
	findFileHandle = FindFirstFile(pattern,&findData);

	//if search wasn't ok, report the error
	if (findFileHandle==INVALID_HANDLE_VALUE )
		result = CResult(codeFail,0,
							"CPackFile::makeFromDir couldn't find any file.");	
	else
	{
		//while we still have more files, continue searching
		while (FindNextFile(findFileHandle,&findData))
		{
			//if the file its a directory
			if (findData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
			{
				//if isn't dir ".."
				if (strstr(findData.cFileName,"..")==NULL)
				{
					//get relative dir name
					strcpy (pattern,directory);
					strcat (pattern, "\\");
					strcat(pattern,findData.cFileName);

					//recursive call
					addDirectory(pattern,basedir);
				}
				
				
			}
			else
			{
				//get the relative filename
				strcpy (pattern,directory);
				strcat (pattern, "\\");
				strcat (pattern,findData.cFileName);

				//add the file
				addFile(pattern,basedir);
			}
		}
		//end of searching
		FindClose(findFileHandle);
	}

	//return value
	return result;
}

CResult CPackFile::addFile(char *fileName,char* basedir)
{
	CResult result = resultOk;


	//increase file number
	header.numberOfFiles++;

	//new file to add to the list
	TFileInfoElement* element = new TFileInfoElement;
	ZeroMemory(element,sizeof(TFileInfoElement));

	//if they were no more files, this is the first
	if (lastFile == NULL)	
		firstFile = element;
	else
	{
		//if they where more files
		if(lastFile!=NULL )
		{
			//insert in the list
			lastFile->next = element;			
			element->prior = lastFile;	
		}
	}

	//set the original file name
	strcpy(element->info.originalFileName,fileName);

	//store relative file name
	char* pointer = fileName;
	pointer += strlen(basedir)+1;
	strcpy(element->info.fileName,pointer);

	//this one is the last one
	lastFile = element;

	return result;
}
CResult CPackFile::finalize()
{
	CResult result = resultOk;

	//call base method
	result = Inherit::finalize();

	//to cicle trougth all elements
	TFileInfoElement* element;

	//delete and detach de element
	while ( lastFile!=NULL)
	{
		element = lastFile;
		lastFile = lastFile->prior;
		delete element;	
	}	

	//return result
	return result;
}

CResult CPackFile::saveFileBody(FILE* outStream,TFileInfo* fileInfo)
{
	CResult result = resultOk;

	//init vars
	char stringBuffer[1024];
	ZeroMemory (stringBuffer,sizeof(stringBuffer));

	//open input file
	FILE* fileInput = fopen(fileInfo->originalFileName,"rb");

	//if couldn't open it
	if (fileInput==NULL)
	{
		//error string
		sprintf(stringBuffer,"CPackFile::saveFileBody - couldn't open file %s",
					fileInfo->originalFileName);
		result = CResult(codeFail,0,stringBuffer);
	}
	else
	{
		//total compress size
		unsigned long int total;

		//get the data manager
		CDataManager manager;
		manager.initialize("");

		//get file pos
		fileInfo->uncompressSize = ftell(fileInput);

		//compress the file
		result = manager.compressFile(fileInput,outStream,header.blocksize,&total);
		
		//store compress size of the file
		fileInfo->compressSize  = total;
		
		//get the file uncompress size
		fseek(fileInput,0,SEEK_END);		
		fileInfo->uncompressSize = ftell(fileInput) -fileInfo->uncompressSize;

		//close the file
		fclose(fileInput);

	}
	
	//return result
	return result;
}

CResult CPackFile::loadFileTable(FILE* inputStream)
{
	CResult result = resultOk;

	//the size of the file table
	unsigned long int tableSize=0;
	
	//steps indicator
	signed long int step;

	//seek from end to read file table size
	step = sizeof(unsigned long int);
	fseek(inputStream, - step ,SEEK_END);	
	fread(&tableSize,1,sizeof(tableSize),inputStream);

	//go to the begining of the table
	step = tableSize+sizeof(unsigned long int);
	fseek(inputStream,-  step ,SEEK_END);
	
	//loops vars
	DWORD counter;
	TFileInfoElement* element = NULL;
	TFileInfo info;
	
	//go for all the files
	for (counter=0;counter<header.numberOfFiles;counter++)
	{

		//new item
		element  = new TFileInfoElement;

		//init with zeroes
		ZeroMemory(element,sizeof(TFileInfoElement));
		ZeroMemory(&info,sizeof(TFileInfo));

		//if theey ar no files this ist the first
		if ( lastFile == NULL)
			firstFile = element;
		else
		{	
			//if they are any
			if(lastFile!=NULL )
			{
				//atach this
				lastFile->next = element;			
				element->prior = lastFile;	
			}
		}	
		
		//this its the last
		lastFile = element;

		//get the file attributes
		fgets(info.fileName,255,inputStream);
		info.fileName[strlen(info.fileName)-1]=0;		
		fread(&info.offset ,1,sizeof(info.offset),inputStream);
		fread(&info.compressSize  ,1,sizeof(info.compressSize),inputStream);
		fread(&info.uncompressSize  ,1,sizeof(info.uncompressSize),inputStream);

		//set the info
		memcpy(&element->info,&info,sizeof(info));
	}

	//return value
	return result;
}

CResult CPackFile::saveFileTable(FILE* outStream)
{
	CResult result = resultOk;

	//to cicle the list
	TFileInfoElement* element = firstFile;

	//current file pos
	unsigned long int initialPosition = ftell(outStream);
	unsigned long int currentPosition;

	//go to all the elements
	while ( (element!=NULL) && (result == resultOk) )
	{	
		//put the file name, offset, compress size & uncompres sizze
		fputs(element->info.fileName,outStream);		
		fputc('\n',outStream);
		fwrite(&element->info.offset ,1,sizeof(element->info.offset),outStream);
		fwrite(&element->info.compressSize  ,1,sizeof(element->info.compressSize),outStream);
		fwrite(&element->info.uncompressSize ,1,sizeof(element->info.uncompressSize),outStream);
	
		element = element->next;		
	}

	//write the size of the file table
	currentPosition = ftell(outStream) - initialPosition;
	fwrite(&currentPosition,1,sizeof(currentPosition),outStream);

	//result value
	return result;
}

CResult CPackFile::saveFilesBodys(FILE* outStream)
{
	CResult result = resultOk;

	//to cicle trougth the list
	TFileInfoElement* element = firstFile;

	//to know file pos
	unsigned long int initialPosition = ftell(outStream);
	unsigned long int currentPosition;

	//go trougth all the files
	while ( (element!=NULL) && (result == resultOk) )
	{
		//now pos
		currentPosition = (ftell(outStream) - initialPosition) + sizeof(header)+1;
		
		//save the file body
		result = saveFileBody(outStream,&element->info);
		
		//set the offset in the file table
		element->info.offset = currentPosition;

		//go to next
		element = element->next;			
	}

	///return value
	return result;
}


CResult CPackFile::loadHeader(FILE* inputStream)
{
	CResult result = resultOk;

	//read the file header
	fread(&header,1,sizeof(header),inputStream);

	//if mark isn't ok, report an error
	if (strcmpi(header.mark,id)!=0)
		result = CResult (codeFail,0,
				"CPackFile::loadHeader input file its not a  pack file.");
	else
	{
	}

	//return value
	return result;
}


CResult CPackFile::saveHeader(FILE* outStream)
{
	CResult result = resultOk;

	//just write it
	fwrite(&header,1,sizeof(header),outStream);

	//result value
	return result;
}

CResult CPackFile::savePackFile(char *fileName)
{
	CResult result = resultOk;

	//open out file
	FILE* outputStream =	fopen(fileName,"wb");

	//if couldn't open it, report a error
	if (outputStream==NULL)
		result = CResult(codeFail,0,"Error couldn't save pack file");
	else
	{
		//save pack header
		result = saveHeader(outputStream);
		
		//if ok
		if (result == resultOk)
		{
			//save the files bodys
			saveFilesBodys(outputStream);

			//if ok save file table
			if (result == resultOk)
				saveFileTable(outputStream);			
		}

		//close out file
		fclose(outputStream);
	}

	//return value
	return result;
}




CResult CPackFile::extractFile(TFileInfo* file, char* directory)
{
	CResult result = resultOk;

	//init vars
	char longName[MAX_PATH];
	char copyName[MAX_PATH];
	char newDir[MAX_PATH];
	ZeroMemory(longName,sizeof(longName));
	ZeroMemory(copyName,sizeof(copyName));
	ZeroMemory(newDir,sizeof(newDir));

	strcpy(longName,directory);
	strcat(longName,"\\");
	strcat(longName,file->fileName);

	//create all dirs to the file, following this format :
	//  temp\
	//  temp\kk
	//  temp\kk\etc
	//  ....
	strcpy(copyName,longName);
	char* p = strstr(copyName,"\\");
	while (p!=NULL)
	{
		*p = 0;

		strcpy(newDir,copyName);
		
		CreateDirectory(newDir,NULL);

		*p = '\\';
		p++;

		p = strstr(p,"\\");
	}

	//streams pointers
	FILE* inputStream;
	FILE* outputStream;

	//open out stream
	outputStream = fopen(longName,"wb");

	//if couldn't open report an error
	if (outputStream==NULL)
		result = CResult(codeFail,0,"Couldn't open output file");
	else
	{
		//open input stream
		inputStream = fopen(fileName,"rb");

		//if couldn't open it, report an error
		if (inputStream==NULL)
			result = CResult(codeFail,0,"Couldn't open input file");
		else
		{
			//get data manager
			CDataManager manager;
			manager.initialize("");


			//go to the offset in the pack
			fseek(inputStream,file->offset-1,SEEK_SET);

			//extrac all file
			result = manager.uncompressFile(inputStream,outputStream,header.
				blocksize,file->compressSize);
			
			//close the stream
			fclose(inputStream);
		}

		//close the stream
		fclose(outputStream);
	}

	//return value
	return result;
}

CResult CPackFile::extractToDir(char *directory)
{
	CResult result = resultOk;


	//cycle trougth all the elements
	TFileInfoElement* element;	
	element = firstFile;

	//until we are in the end
	while ( (element!=NULL) && (result == resultOk) )
	{
	
		//extract the file
		result = extractFile(&element->info,directory);

		//go to next
		element = element->next;				
	}

	//return result
	return result;
}
}