// physicalfile.h: interface for the CPhysicalFile class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(PHYSICALFILE__INCLUDED_)
#define PHYSICALFILE__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "DataFile.h"
#include "stdio.h"

namespace FileSystem
{
//!  Physycal file class
/*!
	Physycal file class. This class its the implemetation of a physicall
	file. 

	\ingroup Base
	\sa CCompressFile
	\sa CDataFile
*/
class CPhysicalFile : public CDataFile  
{
public:

	public : 				
	//! CDataFile class it's the parent of CPhysicalFile class
	/*!
		CDataFile class it's the parent of CPhysicalFile class
	*/
	typedef CDataFile Inherit;

	//!  Default constructor. 
	/*!
		Called before the object creation.
	*/
	CPhysicalFile():Inherit(){};

	//!  Default destructor. 
	/*!
		Called before the object destruction.
	*/
	virtual ~CPhysicalFile(){};

	//!  Initialize an CPhysicalFile Object.
	/*!
		This method initialize the CPhysicalFile object.
		\param fileName = file to be opened
	*/
	virtual CResult initialize(CDataManager* pManager,char *fileName);

	//!  Finalize an CPhysicalFile Object.
	/*!
		This method finalize the CPhysicalFile object, olso close the file.
	*/
	virtual CResult finalize();

	//! Seek an amout of byte in the file
	/*!
		Seek an amout of byte in the file
		\param amout = amout of bytes to be seek
		\pos = SEEK_CUR, SEEK_END, SEEK_SET
		\retval 0 if its correct
	*/
	int seek(long amount,int pos);


	//! Read an amout of byte in the file
	/*!
		Read an amout of byte in the file
		\param amout = amout of bytes to read
		\param buffer* = destination buffer
		\retval number of bytes readed
	*/
	int read(long amount,char* buffer);

	//! To Know if we are in the end of the file
	/*!
		To Know if we are in the end of the file
		\retval if we are at the end of the file
	*/
	int eof();

	//!  Read a string into the buffer
	/*!
		You should call this function to read a
		string into a buffer

		\param maxLen = maximum string size.
		\param buffer = a pointer of a buffer to
		 be fille with the data		 
		 \retval the string itself or NULL if fail
	*/
	char* gets(char* buffer,int maxLen);
private:

	//! its file open
	/*!
	    its file open
	*/
	bool isOpen;

	//! file handle
	/*!
	    file handle
	*/
	FILE* fileHandle;
};

}

#endif // !defined(PHYSICALFILE__INCLUDED_)

